from typing import Any, Dict, List, Optional

from pydantic import parse_obj_as

from plurk.async_apis.base import BaseApi
from plurk.exceptions import validate_resp
from plurk.models import ActionResp, PublicUserData
from plurk.utils import build_params


class FriendsFans(BaseApi):
    async def get_friends_by_offset(self, user_id: int, offset: Optional[int] = None, limit: int = 10):
        """Returns `user_id`'s friend list.

        User can toggle their friend list visibility in privacy settings. If the current user
        can't see the user's friend list, the result will be an empty list.
        """
        endpoint = f'{self.client.base_url}/APP/FriendsFans/getFriendsByOffset'
        params = build_params(user_id=user_id, offset=offset, limit=limit)
        resp = await self.client.http_client.get(endpoint, params=params)
        validate_resp(resp)
        return parse_obj_as(List[PublicUserData], resp.json())

    async def get_fans_by_offset(self, user_id: int, offset: Optional[int] = None, limit: int = 10):
        """Returns `user_id`'s fan list.

        User can toggle their fan list visibility in privacy settings. If the current user
        can't see the user's fan list, the result will be an empty list.
        """
        endpoint = f'{self.client.base_url}/APP/FriendsFans/getFansByOffset'
        params = build_params(user_id=user_id, offset=offset, limit=limit)
        resp = await self.client.http_client.get(endpoint, params=params)
        validate_resp(resp)
        return parse_obj_as(List[PublicUserData], resp.json())

    async def get_following_by_offset(self, offset: Optional[int] = None, limit: int = 10):
        """Returns current user's following list.
        """
        endpoint = f'{self.client.base_url}/APP/FriendsFans/getFollowingByOffset'
        params = build_params(offset=offset, limit=limit)
        resp = await self.client.http_client.get(endpoint, params=params)
        validate_resp(resp)
        return parse_obj_as(List[PublicUserData], resp.json())

    async def become_friend(self, friend_id: int):
        """Create a friend request to `friend_id`. User with `friend_id` has to accept a friendship.
        """
        endpoint = f'{self.client.base_url}/APP/FriendsFans/becomeFriend'
        payload = build_params(friend_id=friend_id)
        resp = await self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())

    async def remove_as_friend(self, friend_id: int):
        """Remove friend with ID `friend_id`. `friend_id` won't be notified.
        """
        endpoint = f'{self.client.base_url}/APP/FriendsFans/removeAsFriend'
        payload = build_params(friend_id=friend_id)
        resp = await self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())

    async def become_fan(self, fan_id: int):
        """Become fan of fan_id.
        """
        endpoint = f'{self.client.base_url}/APP/FriendsFans/becomeFan'
        payload = build_params(fan_id=fan_id)
        resp = await self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())

    async def set_following(self, user_id: int, follow: bool):
        """Update following of user_id. A user can befriend someone, but can unfollow them.
        This request is also used to stop following someone as a fan.
        """
        endpoint = f'{self.client.base_url}/APP/FriendsFans/setFollowing'
        payload = build_params(user_id=user_id, follow=follow)
        resp = await self.client.http_client.post(endpoint, json=payload)
        validate_resp(resp)
        return ActionResp(**resp.json())

    async def remove_as_fan(self, user_id: int):
        """Stop being a fan of the target user.

        The function is shortcut using the setFollowing endpoint, not mapping to a standalone endpoint.
        """
        return await self.set_following(user_id, follow=False)

    async def get_completion(self) -> Dict[str, Dict[str, Any]]:
        """Returns a JSON object of the logged in users friends (nick name and full name).
        This information can be used to construct auto-completion for private plurking.
        Notice that a friend list can be big, depending on how many friends a user has,
        so this list should be lazy-loaded in your application.

        Example returning data:
        {
            '123': {
                'full_name': 'Plurk User',
                'nick_name': 'plurk',
                'avatar': 12345,
                'display_name': 'Plurker'
            }
        }

        where '123' is the user id.
        """
        endpoint = f'{self.client.base_url}/APP/FriendsFans/getCompletion'
        resp = await self.client.http_client.get(endpoint)
        validate_resp(resp)
        return resp.json()



from typing import Dict, TypeVar

from authlib.integrations.httpx_client import AsyncOAuth1Client

from plurk import async_apis, async_oauth
from plurk.clients.base import BaseClient
from plurk.exceptions import validate_resp


T = TypeVar('T')


class AsyncClient(BaseClient):
    """Asynchronous client for Plurk API
    """
    @property
    def http_client_class(self):
        return AsyncOAuth1Client

    async def __aenter__(self):
        await self.setup_client()
        return self

    async def __aexit__(self, exc_type, exc_value, exc_traceback):
        await self.http_client.aclose()

    async def setup_client(self):
        if self.http_client:
            await self.http_client.aclose()  # type: ignore
        self.http_client = async_oauth.get_oauth_client(
            self.app_key,
            self.app_secret,
            self.token,
            self.token_secret,
        )

    async def get_request_token(self):
        return await async_oauth.get_request_token(
            self.http_client,
            request_token_url=f'{self.base_url}/OAuth/request_token',
        )

    def get_auth_url(self, request_token: Dict):
        return async_oauth.get_auth_url(
            self.http_client,
            authenticate_url=f'{self.base_url}/OAuth/authorize',
            request_token=request_token
        )

    async def fetch_access_token(self, request_token: Dict, oauth_verifier: str):
        access_token = await async_oauth.fetch_access_token(
            self.app_key, self.app_secret,
            access_token_url=f'{self.base_url}/OAuth/access_token',
            request_token=request_token,
            oauth_verifier=oauth_verifier,
        )
        self.token = access_token['oauth_token']
        self.token_secret = access_token['oauth_token_secret']
        await self.setup_client()
        return access_token

    async def set_access_token(self, token: str, token_secret: str):
        self.token = token
        self.token_secret = token_secret
        await self.setup_client()

    async def checkToken(self):
        endpoint = f'{self.base_url}/APP/checkToken'
        resp = await self.http_client.post(endpoint)
        validate_resp(resp)
        return resp.json()

    async def expireToken(self):
        endpoint = f'{self.base_url}/APP/expireToken'
        resp = await self.http_client.post(endpoint)
        validate_resp(resp)
        return resp.json()

    async def checkTime(self):
        endpoint = f'{self.base_url}/APP/checkTime'
        resp = await self.http_client.post(endpoint)
        validate_resp(resp)
        return resp.json()

    async def echo(self, data: Dict[str, T]) -> Dict[str, T]:
        endpoint = f'{self.base_url}/APP/echo'
        resp = await self.http_client.post(endpoint, data=data)
        validate_resp(resp)
        return resp.json()

    @property
    def users(self):
        return async_apis.Users(self)

    @property
    def profile(self):
        return async_apis.Profile(self)

    @property
    def realtime(self):
        return async_apis.Realtime(self)

    @property
    def timeline(self):
        return async_apis.Timeline(self)

    @property
    def responses(self):
        return async_apis.Responses(self)

    @property
    def friends_fans(self):
        return async_apis.FriendsFans(self)

    @property
    def alerts(self):
        return async_apis.Alerts(self)

    @property
    def plurk_search(self):
        return async_apis.PlurkSearch(self)

    @property
    def user_search(self):
        return async_apis.UserSearch(self)

    @property
    def emoticons(self):
        return async_apis.Emoticons(self)

    @property
    def blocks(self):
        return async_apis.Blocks(self)

    @property
    def cliques(self):
        return async_apis.Cliques(self)

    @property
    def helpers(self):
        return async_apis.Helper(self)



from abc import abstractmethod


class BaseClient():
    """Base class for clients
    """
    @property
    @abstractmethod
    def http_client_class(self):
        pass

    def __init__(self, app_key: str, app_secret: str, base_url='https://www.plurk.com'):
        self.http_client = None
        self.app_key = app_key
        self.app_secret = app_secret
        self.base_url = base_url
        self.token = None
        self.token_secret = None
